"""
# inovopy
This library provide simple api for application interacting with
inovo robot arm.

## Modules
- `socket`, tcp socket-based communication api
- `geometry`, joint-space and 3D spatial data structure and api
- `iva`, message generation for iva protocal (a robot protocal interfacing with inovo robot arm)
- `rosbridge`, rosbridge API to control psu
- `util`, utilility module
"""
